;***************************************************************************
;* TINYFOX 80m
;* 
;* File Name		: "Tinyfox20b.asm"
;* Title				: TINYFOX 80m
;* Author			: Matthias Kuehlewein, DL3SDO
;* Date				: 12.1.2008
;* Version			: 2.0b
;* Target MCU		: ATtiny13
;* Fuses          : Default, except BODLEVEL1 = 0 => VBOT = 2.7V
;* Compiler			: AVRASM2 (2.0.28) contained in aStudio4b401.exe
;*
;* DESCRIPTION
;* This is the software for the Atmel ATtiny13 on the Tinyfox Foxoring
;* transmitter. 
;* In this compilation the expert mode IS NOT SET as default mode
;* This software version runs internally with 9.6MHz / 16 = 600kHz
;*
;* NOTE
;* This is an alternative 2.0 Version for the ATtiny13. The ATtiny12
;* (V1.X) originally runs at 1.17MHz. The 'official' 2.0 Version runs at
;* 9.6MHz / 8 = 1.2MHz. I generated this version to make tests with
;* the ATtiny13 at 4.8MHz / 8 = 600kHz. Although the power consumption
;* is marginally reduced the deviation of the internally calibrated 
;* clock frequency from the intended 4.8MHz is much higher compared to
;* 9.6MHz. Therefore I went back to 9.6MHz / 8 = 1.2MHz.
;* This version now runs with 9.6MHz / 16 = 600kHz, combining a more
;* precise clock oscillator with an (unknown) supply current reduction.
;* The software is published for information purposes.
;* 
;*
;* Copyright (C) 2005-2008  Matthias Kuehlewein, DL3SDO
;*
;* This program is free software; you can redistribute it and/or modify
;* it under the terms of the GNU General Public License as published by
;* the Free Software Foundation; either version 2 of the License, or
;* (at your option) any later version.
;*
;* This program is distributed in the hope that it will be useful,
;* but WITHOUT ANY WARRANTY; without even the implied warranty of
;* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;* GNU General Public License for more details.
;*
;* You should have received a copy of the GNU General Public License
;* along with this program; if not, write to the Free Software
;* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;*
;***************************************************************************

.INCLUDE "tn13def.inc"					; include made by ATMEL
;The X and Y registers were not available in the ATtiny12 and are therefore
;also not used by this software. To avoid unnecessary warnings, X and Y are 
;undefined below.
.UNDEF XH
.UNDEF XL
.UNDEF YH
.UNDEF YL

;****************************************************************************
;*
;* Definitions and Settings
;*
;****************************************************************************
;****************************************************************************
; Program switches
;****************************************************************************
; Enable Expert Mode as Default Mode
;#define ExpertModeAsDefault

; Always start Key Configuration with the Initial Key (see below), else with
; the stored key
#define ConfigWithInitialKey

; Key Configuration like Delay Configuration:
; Every key is put out several times. If the switch is pressed in time,
; the software proceeds to the next key, else the key is stored
#define ProceedWhenSwitchPressed

;****************************************************************************
; Constant and Parameter Settings
;****************************************************************************
; Programm Version
.EQU Version = 0x20

; Magic Number for Expert Mode
.EQU ExpModeKey = 0x6b

; Initial Key
.EQU InitKey = 0							; 'a'


; Key Speed (smaller values => faster), duration for one dot
; in multiples of 0.0273s
.EQU DotSpeed = 5
#define DotTimes(X) ((X)*DotSpeed)

; Timeout for entering configuration (in ms, range from 0 to 6960ms)
; AVRASM2 doesn't (yet) support floating point arithmetics
.EQU ConfTimeOut = 3500
#define MSTimes(X) ((X)/27)


; during configuration: how often to repeat a key until ALT2: switching to the next
;																		  ALT1: storing the key
#ifdef ProceedWhenSwitchPressed
.EQU CharRepetitions = 6				; for ALT1
#else
.EQU CharRepetitions = 2				; for ALT2
#endif

; how long to wait in expert mode until switching to the key configuration
.EQU WaitForKeyConf = 8

; how often to repeat a delay time until storing the delay value
.EQU RepeatDelay = 5

; how many key repetitions are sent in 1st stage delay until the delay time is put out
.EQU RepeatKey = 4

; Start of Key patterns
.SET SpecChars = (SpecCharsStart-PatternStart)*2
.SET StringChars = (StringCharsStart-PatternStart)*2
.SET CharsEnd = (StringCharsEnd-PatternStart)*2

;****************************************************************************
; Register and Bit Definitions
;****************************************************************************
;***** Working Temporary Register Variables
.DEF temp = r16

;***** Temporary Register Variables
.DEF EEAdr = r17							; used for EEPROM functions
.DEF EEData = r18							; used for EEPROM functions
.DEF BitCounter = r17					; used for CharOut function
.DEF temp2 = r18

;***** Interrupt Variables
.DEF IRQ_statusreg = r10
.DEF IRQ_LWT_Hi = r19
.DEF IRQ_LWT_Lo = r20

;***** Permanent Register Variables
.DEF SpecCharBits = r1					; the number of bits for a spec. char
.DEF Zero = r2								; constant value zero
.DEF StringLen = r3						; number of chars in the string
.DEF Key = r4								; the programmed Key
.DEF StringPos = r5						; position in string
.DEF DelayLo = r6							; delay low byte, granularity 30min, range: 0-9,5h
.DEF DelayHi = r7							; delay high byte, granularity 10h, range: 0-50h
.DEF One = r8								; constant value one
.DEF Delay15 = r9							; delay abs. time, granularity 15min, range: 0-59,5h

.DEF StringOutKey = r22					; the returned OutKey when in string mode
.DEF OutKey = r23							; the key to be output, is set to begin 
												; of next char by CharOut function
.DEF WaitTime = r24						; duration low byte to wait by interrupt routine
												; (decremented in interrupt routine)
.DEF SwitchState = r25					; used to check for released switch

.DEF Control = r26
;***** Bit Definitions for Control Register
.EQU B_LED 			= 0					; only drive LED
.EQU B_TimerRun 	= 1					; set to enable timer, reset by int. routine 
												; when timer elapsed or switch pressed
.EQU B_TXOverride = 2					; TX is overridden by jumper setting
.EQU B_CheckSwitch= 3					; Check for Config switch
.EQU B_TXToSet 	= 4;FIXED!			; Drive Transmitter (is equal to LED postion in PORTB)
.EQU B_Switch 		= 5					; Config. switch pressed (always reset in Wait function)
.EQU B_String 		= 6					; set when processing a string

;***** Bit Definitions for IO Registes
.EQU B_IOTX 		= 3
.EQU B_IOLED 		= 4

.DEF RepChar = r27						; used for key repetitions

.DEF Control2 = r28
;***** Bit Definitions for Control2 Register
.EQU B_DelTimRun = 0						; expert mode: start delay timer, 
												; is reset by interrupt routine when elapsed
.EQU B_30TestCycle = 1					; expert mode: jump back to 30 min test cycle loop

.DEF AddWait = r29						; additional time to wait in CharOut function

;Unused registers:
; r11-r15,r21


;***** EEPROM layout and definitions *********************************************************
.EQU	EE_Version	= 0					; program version at address 0
.EQU	EE_Key		= 1					; key at address 1
.EQU	EE_OscCal	= 2					; oscillator calibration value at address 2
.EQU	EE_ExpMode	= 3					; expert mode at address 3
.EQU	EE_DelayLo	= 4					; start delay low (valid in expert mode) at address 4
.EQU	EE_DelayHi	= 5					; start delay high (valid in expert mode) at address 5


;****************************************************************************
;*
;* Main Program
;*
;****************************************************************************
;****************************************************************************
; Vector section
;****************************************************************************
		rjmp	RESET							; Reset handler
		reti
		reti
		rjmp	TIM0_OVF						; Timer0 overflow handler
		reti
		reti
		reti
		reti
		reti
		reti
;****************************************************************************
; Init
;****************************************************************************
;some inits
;the stack pointer is initialized although its power-on value is ok, but
;it's unclear what happens after a reset caused by a watchdog or brown-out reset
RESET:
		ldi	temp,LOW(RAMEND)
		out	SPL,temp						; set stack pointer to end of RAM
		ldi	temp,(1<<CS01|1<<CS00)	; CK/64
		out	TCCR0B,temp
		ldi	temp,(1<<TOIE0)			; use timer overflow
		out	TIMSK0,temp
		clr	Control						; init. Control register
		clr	Control2						; init. Control2 register
		clr	Zero							; init. Zero register to zero
		ldi	temp,0x01
		mov	One,temp						; init. One register to one
		clr	AddWait						; init Add. Wait Time register
		clr	IRQ_LWT_Hi
		clr	IRQ_LWT_Lo
		out	PORTB,Zero					; is already zero by default
		out	DDRB,Zero					; is already zero by default
		sbi	ACSR,ACD						; switch off Analog Comparator
		cli									; disable interupts just to be absolutely sure
		ldi	temp,(1<<CLKPCE)
		out	CLKPR,temp					; prepare changing the clock prescale register
		ldi	temp,(1<<CLKPS2)			; divide the 9.6MHz by 16 = 600 kHz
		out	CLKPR,temp					; change the clock prescale register
		sei									; enable interrupts, i.e. the timer interrupt

;check whether the device runs for the first time after being programmed
		clr 	EEAdr							; Version at address 0
		rcall	EERead
		cpi	EEData,Version
		breq	Startup           		; normal Startup
;else program EEPROM
		clr 	EEAdr							; Version at address 0
		ldi	EEData,Version
		rcall	EEWrite
		ldi	EEData,InitKey
		rcall	EEWrite						; initial key at address 1
		in		EEData,OSCCAL
		rcall	EEWrite						; write current Oscillator Calibration Value to address 2
#ifdef ExpertModeAsDefault
#message "Compilation with Expert Mode as default mode"
		ldi	EEdata,ExpModeKey			; enable expert mode by default
#else
		clr	EEData						; disable expert mode
#endif
		rcall	EEWrite						; at address 3
		rcall	EEWriteZero					; set init. delay low part at address 4 to zero
		rcall	EEWriteZero					; set init. delay high part at address 5 to zero

;****************************************************************************
; Startup
;****************************************************************************
Startup:
		sbi	PORTB,PINB0					; enable pull-up at PINB0
		ldi 	EEAdr,EE_Key				; Key at address 1
		rcall	EERead
		mov	Key,EEData
		rcall	EERead						; Oscillator calibration value at address 2
		tst	EEData
		breq	w51							; don't accept zero as calibration value, something
												; seems to have gone wrong
		in		temp,OSCCAL
l50:	cp		temp,EEData
		breq	w51							; finished, oscillator trimmed to target value
		brlo	w52
;OSCCAL higher than target value in EEProm
		subi	temp,2
w52:	inc	temp
		out	OSCCAL,temp
		rjmp	l50
		
w51:	rcall	EERead						; check expert mode byte at address 3
		clt
		cpi	EEData,ExpModeKey
		brne	w50
		set									; set T Flag in SREG to indicate expert mode		
		
; check for pressed Prog.switch
w50:	sbis	PINB,PINB0					; check value at PINB0 (= switch)
		rjmp	CONFIGURE
; switch is not pressed, start normal operation
		clt									; switch is not pressed, enforce normal operation
												; without delay by clearing the T-flag
		
;****************************************************************************
; Entry Point for normal operation (either with or without delay)
;****************************************************************************
e1:	wdr									; reset watchdog before activating it
		ldi	temp,0b00111001			; WDE = 1, WDCE=1, WDP[3..0] = 1001, timeout 8s
		out	WDTCR,temp					; enable watchdog

		cbi	PORTB,PINB0					; disable pull-up at PINB0 (not needed, save power)
		cbr	Control,1<<B_LED			; disable LED indication only
		
		in		temp,MCUSR					; check source of reset
		out	MCUSR,Zero					; reset MCU status register, clear all reset flags
		sbrc	temp,WDRF
		rjmp	s1								; OUCH! a watchdog reset occured, skip delay and go
												; right away to normal operation

		brtc	s1								; no expert mode -> Normal Operation
		
		ldi	EEAdr,EE_DelayLo
		rcall	EERead						; read low part of delay value
		mov	DelayLo,EEData
		rcall	EERead						; read high part of delay value
		mov	DelayHi,EEData
		mov	temp,DelayLo
		or		temp,DelayHi
		breq	s1								; no delay -> Normal Operation
		
;****************************************************************************
; Delay Operation in Expert Mode
;****************************************************************************
; EXPERT MODE
; 1st stage: test cycle execution for 30min
		ldi	temp,2						; 2*15min = 30min
		mov	Delay15,temp
		sbr	Control2,(1<<B_DelTimRun)|(1<<B_30TestCycle)
d1l1:	ldi	RepChar,RepeatKey
d1l2:	mov	OutKey,Key
		cpi	RepChar,1
		brne	d11
		ldi	AddWait,25					; add a pause after the normal key sequence
d11:	rcall	CharOut
		dec	RepChar
		brne	d1l2
		ldi	OutKey,(Correction-PatternStart)*2	; correction tone (8 dots), to indicate subsequent
																; output of delay time
		rcall	CharOut
		rjmp	l46							; put out the programmed delay time
d1l3:	rcall	CharOut						; put out last char of delay
		sbrc	Control2,B_DelTimRun
		rjmp	d1l1

; 2nd stage: wait delay till normal operation
d2:	mov	Delay15,DelayLo
		tst	DelayHi
		breq	d21							; nothing todo. Btw, one could omit this check:
												; if DelayHi is zero, adding any value 256 times
												; leads to the same value in Delay15 as adding never
		ldi	temp,20						; 10h = 20 * 30min
d2l1:	add	Delay15,temp
		dec	DelayHi						; modify DelayHi, it's not needed anymore
		brne	d2l1		
d21:	dec	Delay15						; - 1 since we spent already 30min in the 1st stage
		lsl	Delay15						; * 2 to get the 15min units from the 30min part
		sbr	Control2,1<<B_DelTimRun
d2l2:	wdr									; trigger watchdog
		sbrc	Control2,B_DelTimRun
		rjmp	d2l2							; wait till delay elapses
		
;****************************************************************************
; Normal Operation, no delay or delay elapsed
;****************************************************************************
; this is the main loop during normal operation on the topmost level !
s1:	mov	OutKey,Key
		rcall	CharOut
		rjmp	s1


;****************************************************************************
; Configuration
;****************************************************************************
CONFIGURE:
		sbr	Control,1<<B_LED			; disable TX, only LED indication
		ldi	WaitTime,MSTimes(ConfTimeOut)	; wait for max. the config. time
		ldi	SwitchState,0x03			; set switch state initially
		sbr	Control,(1<<B_TimerRun)|(1<<B_TXToSet)	; start waiting and drive LED to show reaction
		rcall SetTX
l40:	sbrs	Control,B_TimerRun		; wait for operator to release switch
		rjmp	e1								; not released within TIMEOUT -> Entry Point for Delay Jumper
		tst	SwitchState
		brne	l40							; still pressed
		
		rcall SetTXOff						; turn off LED
		ldi	temp,0x0f
		out	PORTB,temp					; activate pull-ups for expert mode checking
		sbr	Control,1<<B_CheckSwitch	; enable switch checking

; check for expert mode switch
		in		temp,PINB
		andi	temp,0x0f
		cpi	temp,0x03					; SCK must be pulled low externally, MISO and MOSI are
												; pulled up by internal pull-up, PB3 is pulled-low by external R
		brne	w43							; no expert mode switch
; don't care for activated pull-ups ...
		ldi 	EEAdr,EE_ExpMode			; Expert Mode at address 3
		clr	EEData
		brts	w44							; device was in expert mode
		ldi	EEData,ExpModeKey
w44:	rcall	EEWrite
		rcall	EEWriteZero					; clear delay values
		rcall	EEWriteZero
		cbr	Control,1<<B_CheckSwitch	; disable switch checking
l44:	ldi	OutKey,17					; 'r'
		rcall	CharOut
		rjmp	l44							; !! endless loop !!
		
; output version to indicate the initialization
w43:	ldi	OutKey,CharsEnd			; output version information (string)
		ldi	AddWait,10					; add a pause after the version string
		rcall	CharOut

		brtc	KeyConfig					; no expert mode, configure key
		ldi	RepChar,WaitForKeyConf	; how long to wait until switching to the key configuration
l45:	ldi	OutKey,4						; 'e'
		ldi	AddWait,25					; add another pause after each 'e'
		rcall	CharOut
		sbrc	Control,B_Switch
		rjmp	KeyConfig					; Switch pressed => go to key configuration
		dec	RepChar
		brne	l45

;*******************
;Delay Configuration
;*******************
		clr	DelayLo
wl47:	clr	DelayHi
l47:	ldi	RepChar,RepeatDelay		; how often to repeat a delay time until
												; storing the delay value
; put out current delay
; is also entry point from Delay Operation in Expert Mode
l46:	tst	DelayHi
		breq	w45
		ldi	OutKey,26					; offset for '0'
		add	OutKey,DelayHi
		rcall	CharOut
		sbrc	Control,B_Switch
		rjmp	IncDelay						; Switch pressed => increment Delay
w45:	ldi	OutKey,26<<1				; 2 * offset for '0'
		add	OutKey,DelayLo
		lsr	OutKey						; because the granularity is 30min
		sbrs	DelayLo,0					; check for 30min, could also be done by checking the carry flag
		rjmp	w46
		rcall	CharOut
		sbrc	Control,B_Switch
		rjmp	IncDelay						; Switch pressed => increment Delay
		ldi	OutKey,(Daaah-PatternStart)*2	; long tone, special 30min indicator
w46:	ldi	AddWait,35					; add another pause after each sequence of chars
		sbrc	Control2,B_30TestCycle
		rjmp	d1l3							; jump back in 1st stage of Delay Operation
		cpse	RepChar,One
		ldi	AddWait,DotTimes(3)		; add another pause (3 dots long) after the complete number
		rcall	CharOut
		sbrc	Control,B_Switch
		rjmp	IncDelay						; Switch pressed => increment Delay
		
		dec	RepChar
		brne	l46

; store actual delay in EEPROM
		ldi 	EEAdr,EE_DelayLo			; DelayLo at address 4
		mov	EEData,DelayLo
		rcall	EEWrite						; EEAdr is incremented by EEWrite
		mov	EEData,DelayHi
		rcall	EEWrite
		rjmp	ConfExit						; go to Key. Conf. Exit below
												; show the progr. delay (although the delay jumper is not set !)
		
IncDelay:
		mov	temp,DelayLo
		inc	DelayLo
		or		temp,DelayHi
		breq	IncDelay						; skip the 30min and go from the 0h to the 1h directly
		ldi	temp,20
		cp		DelayLo,temp
		brne	l47
		clr	DelayLo
		inc	DelayHi
		ldi	temp,6
		cp		DelayHi,temp
		brne	l47
		rjmp	wl47

;*****************
;Key Configuration
;*****************
KeyConfig:
#ifdef ConfigWithInitialKey
		ldi	temp,InitKey					; begin with the default char
		mov	Key,temp
#endif
		
;ALTERNATIVE1: like for the delay, every key is put out several times. If the switch
; is pressed in time, the software proceeds to the next key, else the key is stored.
;ALTERNATIVE2: every key is put out several times. If the switch is pressed in time the
; key is stored, else the software proceeds to the next key.

l41:	ldi	RepChar,CharRepetitions	; how often to repeat a char until
												; - switching to the next (ALT2)
												; - or storing it in the EEPROM (ALT1)
l42:	mov	OutKey,Key
		cpi	RepChar,1
		brne	w41
		ldi	AddWait,35					; add another pause after each sequence of chars
w41:	rcall	CharOut
		sbrc	Control,B_Switch
#ifdef ProceedWhenSwitchPressed
		rjmp	w42							; ALT1: Switch pressed => proceed to next key
#else
		rjmp	w40							; ALT2: Switch pressed => store key
#endif
		dec	RepChar
		brne	l42
#ifdef ProceedWhenSwitchPressed
		rjmp	w40							; ALT1: Switch not pressed in time => store key
#endif
w42:	mov	Key,OutKey
		rjmp	l41

w40:	ldi 	EEAdr,EE_Key				; Key at address 1
		mov	EEData,Key
		rcall	EEWrite						; EEAdr is incremented by EEWrite
		ldi 	EEAdr,EE_Key				; Key at address 1 (because EEAdr was 
												; incremented by EEWrite above)
		rcall	EERead						; control success
		mov	Key,EEData
		clt									; ignore delay, since no jumper is set !
ConfExit:
		cbr	Control,1<<B_CheckSwitch	; disable switch checking
		ldi	WaitTime,MSTimes(3500)	; add another longer pause (3.5s) to indicate programming
		rcall	Wait
		rjmp	e1								; jump to Entry Point to start operation

;****************************************************************************
; Timer interrupt service routine
;****************************************************************************
; This function is called after 256 (Timer Register Overflow) * 64 (Clock Prescaler)
; oscillator ticks
; The values are based on the assumption of an calibrated ATtiny13 at 3.3V that runs
; at 600kHz (internal RC oscillator with nominal 9.6MHz and a 16 divider)
; That results in an invocation of this function of round about 0.0273s
; Short delays are therefore selectable between 0.0273s and 6.96s
; The 0x80bf for 15min results from 15 * 60s * 600KHz / (256 * 64)
TIM0_OVF:
		in		IRQ_statusreg,SREG		; save Status register
		tst	SwitchState
		breq	w22
		sbis	PINB,PINB0					; check switch
		ldi	SwitchState,0x03			; still pressed -> set switch state (+1 for dec below)
		dec	SwitchState
w22:	sbrs	Control,B_TimerRun
		rjmp	w20							; no running short delay timer
		dec	WaitTime
		brne	w20
		cbr	Control,1<<B_TimerRun

w20:	sbrs	Control2,B_DelTimRun
		rjmp	w24							; no running long delay timer
		tst	Delay15
		breq	w25
		add	IRQ_LWT_Lo,One
		adc	IRQ_LWT_Hi,Zero
		cpi	IRQ_LWT_Lo,0xbf
		brne	w24
		cpi	IRQ_LWT_Hi,0x80
		brne	w24
		dec	Delay15
		brne	w26
w25:	cbr	Control2,1<<B_DelTimRun
w26:	clr	IRQ_LWT_Lo
		clr	IRQ_LWT_Hi
w24:	out	SREG,IRQ_statusreg		; restore Status register
		reti

;****************************************************************************
; EEPROM service routines, 1st stack level
;****************************************************************************
EERead:
		sbic 	EECR,EEPE					; if EEPE not clear
		rjmp	EERead						; wait more
		out 	EEARL,EEAdr					; set EEPROM address
		sbi	EECR,EERE					; set EEPROM Read strobe
		in		EEData,EEDR					; get data
		rjmp	eeo

EEWriteZero:								; entry point to write a zero to the EEPROM
		clr	EEData
EEWrite:										; normal entry point
		cli									; disable interrupts before writing to EEPROM
el1:	sbic	EECR,EEPE					; wait for EEPE to be cleared
		rjmp	el1							; just for safety, since we wait also at the end
		out 	EEARL,EEAdr					; set EEPROM address
		out 	EEDR,EEData					; set EEPROM data
		sbi 	EECR,EEMPE					; set master program enable
		sbi	EECR,EEPE					; set EEPROM Program strobe
el2:	sbic	EECR,EEPE					; wait for EEPE to be cleared
		rjmp	el2
		sei									; enable interrupts again
eeo:	inc	EEAdr							; increment address for subsequent write (or read) operations
		ret


;****************************************************************************
; Character output routines, 1st stack level
;****************************************************************************
ProcString:
		sbrc	Control,B_String			; are we already in string mode ?
		rjmp	nchar							; ignore
		sbr	Control,1<<B_String		; set string mode
		mov	StringLen,R0
		mov	StringOutKey,OutKey
		add	StringOutKey,StringLen
		cpi	StringOutKey,CharsEnd
		brlo	w7
		clr	StringOutKey				; set to first normal char ('a')
w7:	mov	StringPos,OutKey			; keep track of actual char

w8:	ldi	ZH,HIGH(PatternStart*2)
		ldi	ZL,LOW(PatternStart*2)
		add	ZL,StringPos
		adc	ZH,Zero
		lpm									; -> R0
		mov	OutKey,R0
		rjmp	CharOutInt

;if key was not pressed
cout:
		sbrs	Control,B_String			; are we in string mode
		rjmp	w9								; no, return
nchar:
		inc	StringPos
		dec	StringLen
		brne	w8
; finished string
		mov	OutKey,StringOutKey
		ldi 	temp,DotTimes(3)			; additional pause after a word (3 dots long)
		add	AddWait,temp
; check if additional wait time is set
w9:	tst	AddWait
		breq	o1								; if zero -> exit function
		mov	WaitTime,AddWait
		clr	AddWait
		rcall Wait							; don't check switch, return anyway
o1:	ret

;if key was pressed
kout:
		sbrc	Control,B_String			; are we in string mode
		mov	OutKey,StringOutKey		; yes, update OutKey
		ret

; entry point from topmost program level
CharOut:
		cbr	Control,1<<B_String		; clear string mode
; entry point from this level
CharOutInt:
		ldi	ZH,HIGH(PatternStart*2) ; * 2, since flash is adressed bytewise
		ldi	ZL,LOW(PatternStart*2)
		add	ZL,OutKey
		adc	ZH,Zero
		lpm									; -> R0
		mov	temp,OutKey
		inc	OutKey
		tst	R0
		brne	w6
		cpi	OutKey,CharsEnd
		brlo	CharOutInt					; proceed to next memory location
		clr	OutKey
		rjmp	CharOutInt					; proceed to next memory location
w6:	cpi	temp,StringChars
		brsh	ProcString
		cpi	temp,SpecChars
		brsh	ProcSpecChars
		ldi	BitCounter,0x08
l2:	dec	BitCounter
		breq	CharOutInt					; empty char -> proceed to next one
												; don't put empty chars inside strings !
		lsl	R0
		brcc	l2
w1:	sbr	Control,1<<B_TXToSet		; turn on TX
		rcall SetTX
		ldi	WaitTime,DotTimes(3)		; a dash (= 3 dots long)
		lsl	R0
		brcs	w2
		ldi 	WaitTime,DotTimes(1)		; a dot
w2:	rcall Wait
		sbrc	Control,B_Switch
		rjmp	kout							; config switch pressed -> exit function
		rcall SetTXOff						; turn off TX
		ldi	WaitTime,DotTimes(1)		; wait one dot long between the dots and dashes of a char
		dec	BitCounter
		brne	w3
		ldi	WaitTime,DotTimes(3)		; wait 3 dots long after a char
w3:	rcall Wait
		sbrc	Control,B_Switch
		rjmp	kout							; config switch pressed -> exit function
		tst	BitCounter
		breq	cout							; all bits processed -> exit normal bit processing
		rjmp	w1

ProcSpecChars:
		clr 	BitCounter
		mov	SpecCharBits,R0
		dec	R0
		lsr	R0
		lsr	R0
		lsr	R0
		add	OutKey,R0
		inc	OutKey
l3:	ldi	temp,0x07
		and	temp,BitCounter
		brne	w4
		add	ZL,One
		adc	ZH,Zero
		lpm
w4:	cbr	Control,1<<B_TXToSet
		lsl	R0
		brcc	w5
		sbr	Control,1<<B_TXToSet
w5:	rcall SetTX
		ldi	WaitTime,DotTimes(1)		; wait one dot long, this is the base unit
		rcall Wait
		sbrc	Control,B_Switch
		rjmp	kout							; config switch pressed -> exit function
		inc	BitCounter
		cp		BitCounter,SpecCharBits
		brne	l3
		rjmp	cout							; all bits processed -> exit spec. bit processing

;****************************************************************************
; Wait, 1st or 2nd stack level
;****************************************************************************
Wait:
		wdr									; always reset watchdog
		cbr	Control,1<<B_Switch		; always reset switch flag
		sbr	Control,1<<B_TimerRun
l30:	sbrs	Control,B_TimerRun
		ret									; time elapsed
		sbrc	Control,B_CheckSwitch
		cpse	SwitchState,Zero			; was cleared in interrupt service routine ?
												; the line above is questionable
		rjmp	l30							; don't check switch or switch state was not 
												; yet cleared in int. serv routine
		sbic	PINB,PINB0					; check switch
		rjmp	l30							; switch is not pressed, keep on waiting
		clr	AddWait						; reset add. wait time
		sbr	Control,(1<<B_Switch)|(1<<B_TXToSet)	; set switch flag and turn on TX
		ldi	SwitchState,0x02			; set switch state
		cli									; because we are already at 2nd stack level
		rcall SetTX
		sei
l31:	tst	SwitchState					; check for switch state (will be reset in int. routine)
		brne	l31
		rcall SetTXOff						; turn off TX, no irq locking necessary, 
												; since int. routine was executed just before
		ldi	WaitTime,30					; add a pause before doing anything else
		sbr	Control,1<<B_TimerRun
l32:	sbrc	Control,B_TimerRun
		rjmp	l32
		ret
		
;****************************************************************************
; Drive LED and TX, 1st or 2nd stack level
;****************************************************************************
; PB0 = config. switch, input with selectable internal pull-up
; PB3 = tx,  tristate if disabled, push-pull one for transmitting
; PB4 = led, tristate if disabled, push-pull zero for lighing led (=normally when transmitting)
SetTXOff:									; entry point to turn off the TX
		cbr	Control,1<<B_TXToSet		; turn off TX
SetTX:										; entry point to derive the TX setting from B_TXToSet
		sbrc	Control,B_LED
		rjmp	w11							; only LED
		sbic	DDRB,B_IOTX					; check for TX override by user while TX is input
		rjmp	w10
		cbr	Control,1<<B_TXOverride
		sbic	PINB,B_IOTX
		sbr	Control,1<<B_TXOverride
w10:	sbrc	Control,B_TXToSet
		rjmp	TXOn
;TXOff
		ldi	temp,0x00					; 0x01 for config. switch pull-up (as in version 1.0)
		ldi	temp2,0x10					; prepare for LED indication in case of TX override
		sbrc	Control,B_TXOverride
		rjmp	w13
w12:	out	DDRB,Zero					; switch off: first set to inputs
		out	PORTB,temp					; then set pull-ups
		ret
TXOn:	ldi	temp,0x08					; 0x09 for config. switch pull-up (as in version 1.0)
		ldi	temp2,0x18
w13:	out	PORTB,temp					; switch on: first prepare outputs
w14:	out	DDRB,temp2					; then activate output
		ret
;only LED
w11:	mov	temp2,Control
		andi	temp2,0x10
		rjmp	w14

;****************************************************************************
; Key patterns
;****************************************************************************
;normal chars
PatternStart:
.DB 0b00000101 , 0b00011000		; a, b		0
.DB 0b00011010 , 0b00001100		; c, d		2
.DB 0b00000010 , 0b00010010		; e, f		4
.DB 0b00001110 , 0b00010000		; g, h		6
.DB 0b00000100 , 0b00010111		; i, j		8
.DB 0b00001101 , 0b00010100		; k, l		10
.DB 0b00000111 , 0b00000110		; m, n		12
.DB 0b00001111 , 0b00010110		; o, p		14
.DB 0b00011101 , 0b00001010		; q, r		16
.DB 0b00001000 , 0b00000011		; s, t		18
.DB 0b00001001 , 0b00010001		; u, v		20
.DB 0b00001011 , 0b00011001		; w, x		22
.DB 0b00011011 , 0b00011100		; y, z		24
.DB 0b00111111 , 0b00101111		; 0, 1		26
.DB 0b00100111 , 0b00100011		; 2, 3		28
.DB 0b00100001 , 0b00100000		; 4, 5		30
.DB 0b00110000 , 0b00111000		; 6, 7		32
.DB 0b00111100 , 0b00111110		; 8, 9		34
.DB 0b00110001 , 0					; =			36

;special chars
SpecCharsStart:
Daaah:
.DB 12 , 0b01111111
.DB 0b00000000 , 0b00000000
Correction:
.DB 20 , 0b10101010
.DB 0b10101010 , 0b00000000

;strings
StringCharsStart:
.DB 3 ,12								; len=3 , m
.DB 14,4									; o e

.DB 3 ,12								; len=3 , m
.DB 14,8									; o i

.DB 3 ,12								; len=3 , m
.DB 14,18								; o s

.DB 3 ,12								; len=3 , m
.DB 14,7									; o h

.DB 3 ,12								; len=3 , m
.DB 14,31								; o 5

.DB 2 ,12								; len=2 , m
.DB 14, 4								; o , len=4
.DB 12,14								; m , o
.DB 4 , (Daaah-PatternStart)*2	; e , daaah

.DB 4, 12								; len=4 , m
.DB 14,8									; o i
.DB (Daaah-PatternStart)*2,0		; daaah, <filler byte (is ignored)>

.DB 4, 12								; len=4 , m
.DB 14,18								; o s
.DB (Daaah-PatternStart)*2,0		; daaah, <filler byte (is ignored)>

.DB 4, 12								; len=4 , m
.DB 14,7									; o h
.DB (Daaah-PatternStart)*2,0		; daaah, <filler byte (is ignored)>

.DB 4, 12								; len=4 , m
.DB 14,31								; o 5
.DB (Daaah-PatternStart)*2,0		; daaah, <filler byte (is ignored)>

.DB 3, 12								; len=3 , m
.DB 14,(Daaah-PatternStart)*2		; o daaah

StringCharsEnd:
;everything below the StringCharsEnd label is not selectable as key
;
;version string
.DB 5, 21								; len=5 , v
.DB 28,26								; 2 0
.DB 1,(Daaah-PatternStart)*2		; b daaah


;****************************************************************************
; Changelog
;****************************************************************************
;
;   Date   Version   Change
;----------------------------------------------------------------------------
; 23.01.05   1.0     Release
; 07.08.05   1.1     - The TX is switched off during configuration since the
;                      reading of the config. switch is not always immune against EMC
;                    - The PB0 pull-up was always enabled. Since the config.
;                      switch is now only checked when the TX is off, the
;                      pull-up is disabled after the initial checks for configuration
;                    - added some comments
;                    - modified some comments
; 26.05.06           The software is ported to the ATtiny13
; 12.01.08   2.0b    Release for information purpose (see NOTE section at top of the file)
;